/*------------------------------------------------------------------------------
 *   PERF_METER.C - Performance Meter display for DECwindows Motif, OpenVMS
 *
 *   History:
 *
 *	V01-001		Robert Heller			08-Nov-1989
 *		Created on Wed Nov  8 10:39:17 1989
 *	V01-002		Mark Pizzolato			22-Nov-1989
 *		mark@infopiz.uucp (uunet!lupine!infopiz!mark)
 *	V01-003		Robert Heller			27-Nov-1989
 *		Fixed for invert displays
 *		Modified to use set_label in place of certain set_something
 *		calls and new version of init_cpu_times.
 *
 *	V01-004		Phil Ottewell/Tony Davis	19-OCT-1992
 *		Modified to use UIL/Motif toolkit instead of DECwindows
 *
 *	V02-000		Phil Ottewell			27-NOV-1992
 *		Modified to use different colours for CPU modes,
 *		use a decent icon, and allow more user configuration.
 *
 *	V02-001		Joe Kazimierczyk (kaz@bms.com)	14-FEB-1993 
 *		Plugged memory leak caused by calls to DwtLatin1String without
 *		corresponding XtFree().
 *		Added 'top process' info to bottom of main window
 *		Added toggle button to create a 'show process/cont' type
 *		display for the current top process.
 *		Moved exit button to a popup menu instead of the menu bar,
 *		to save display space.
 *
 *	V02-000-A	Phil Ottewell			20-JAN-1993
 *		Modified V2.0 to work on Alpha
 *
 *	V02-002		Phil Ottewell			01-MAY-1995
 *		Modified to work with DEC C and Motif 1.2
 *		without having to specify /STANDARD=VAXC
 *		Merged Joe Kazimierczyk V2.1 changes in to V02-000-A
 *
 *      v02-002-a       Dean Schamberger                20-may-1995
 *              Modified to protect against setting update interval to 0 secs.
 *              Also updated .UIL to allow up to 4 devices again,
 *              and set the default to a blank line.  Added fix to device
 *              name string to keep columns aligned, and forced TOPcpu
 *              process to stay a fixed length.  This fixed some strange
 *              redraws on some x-terminals.  Finally added the option to
 *              compile a non-priv version that can still see CPU users
 *              (ie.  does not need CMKRNL privs like in version 2.0).
 *              The default is not to need CMKRNL.  To return to the
 *		previous mode, comment out the define NON_PRIV_USERS.
 *		(no longer quite true - see V02-003 below - P.O.)
 *		This only effects CPU states and usage, one still needs
 *		privs to see max cpu user.
 *	V02-003		Phil Ottewell			01-MAY-1995
 *		Fix memory leak caused byXmStringLtoRCreate (Motif pre 1.2-3)
 *		by using XmStringCreateSimple instead. The NON_PRIV_USER option
 *		has been taken out, and the VAX version builds without calling
 *		sys$cmkrnl, whereas the Alpha version HAS to get the CPU info
 *		with kernel mode access. An informational message to this effect
 *		is now issued by the MAKE_PERF_METER.COM procedure.
 *----------------------------------------------------------------------------*/

/* ANSI C headers */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* VMS headers */
#include <descrip.h>
#include <dvidef.h>
#include <jpidef.h>
#include <ssdef.h>
#include <starlet.h>
#include <syidef.h>

/* Local headers */
#ifdef __ALPHA
#include "cpudef_alpha.h"
#else
#include "cpudef.h"
#endif
#include "perf_meter_const.h"

/* X Toolkit header */
#include <X11/Intrinsic.h>
/* Motif general header (includes most Xlib stuff too) */
#include <Xm/Xm.h>
/* Other Motif headers */
#include <Xm/RowColumn.h>
/* Motif Resource Manager*/
#include <Xm/MrmAppl.h>

/* #defines go here */

#define PROGRAM_TITLE_FORMAT "System Performance Meters on %s" /* nodename */
#define PROGRAM_NAME         "PerfMeter"
#define PROGRAM_CLASS        "Meter"
#define PROGRAM_VERSION      "2.3"

#define MX_STATES 7

#define PARAM_STRING 0
#define PARAM_INTEGER 1
#ifdef __ALPHA
#define K_TIME 0
#define E_TIME 2
#define S_TIME 4
#define U_TIME 6
#define I_TIME 8
#define C_TIME 0
#define SPIN_TIME 10
#define IDLE_TIME 12
#else
#define K_TIME 0
#define E_TIME 1
#define S_TIME 2
#define U_TIME 3
#define I_TIME 4
#define C_TIME 5
#define SPIN_TIME 6
#define IDLE_TIME 7
#endif

#define CompTimeX  2
#define NullTimeX 12
#define UserTimeX 22
#define SupeTimeX 32
#define ExecTimeX 42
#define KernTimeX 52
#define InteTimeX 62

#define StateBarWidth 7

/* Global references */
globalref long   EXE$GL_ABSTIM_TICS;
globalref long   SMP$GL_ACTIVE_CPUS;
globalref $CPU*  SMP$GL_CPU_DATA[1];
globalref long   PMS$GL_FAULTS;
globalref long   PMS$GL_PREADS;
globalref long   PMS$GL_PREADIO;
globalref long   PMS$GL_PWRITES;
globalref long   PMS$GL_PWRITIO;
globalref long   PMS$GL_DZROFLTS;
globalref long   SCH$GL_FREECNT;
globalref long   SCH$GL_MFYCNT;
#define CURRENT_FAULTS     (PMS$GL_FAULTS)
#define CURRENT_PAGEIO     (PMS$GL_PREADIO + PMS$GL_PWRITIO)
#define CURRENT_FREEPAGES  (SCH$GL_FREECNT)
#define CURRENT_MODPAGES   (SCH$GL_MFYCNT)


/* Function prototypes */

void init_application(void);

void s_error(char* problem_string);
void get_something(Widget w, char* resource, XtArgVal value);
void set_something(Widget w, char* resource, XtArgVal value);

void activate_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s);
void create_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s);
void cancel_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s);
void update( XtPointer data, XtIntervalId *id );

void create_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s);
void exposure_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s);
void update_contproc( XtPointer data, XtIntervalId *id );
void toggle_contproc(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *toggle);
void contproc_on(void);
void contproc_off(void);
void contproc_jpi_status(int status);
void doMainPopupMenu(Widget wdgt,XtPointer client_data,
                     XEvent *event, Boolean *propogate);
void DrawStateGraph(void);

int initialize_topproc(void);
int get_topproc(void);
char *format_topproc(int top);
int topproc_getPID(int top);

void format_cputime (int t, char *s);

void get_initial_kernelly_stuff();
void get_new_kernelly_stuff();

void init_cpu_times();
void get_cpu_times(long* null, long* user, long* super, long* exec,
                   long* kernel, long* interrupt, long* compat);
void get_disk_opcnt(register long devnum, register long* count);
void update_disk_name_label(Widget w, long devnum);

GC MakeGC(Display* disp, Window wind, long fun, Pixel foreground);

/* Macros */
#define FAILED(status) (((status)&1) == 0)
#define SIGNAL_ERROR(status)  \
    {                         \
      int ___status = status; \
      if (FAILED(___status))  \
      lib$signal(___status);  \
    }

#define MAX_WIDGETS (MAX_WIDGET + 1)

/* Macro to load up an element of a resource list for a resource enquiry */
#define PFM_LOAD_RESOURCE(res,rname,rclass,rtype,rsize,roffset,\
                          defrtype,defraddr) \
{ \
XtResource *_resptr = &(res); /* Only dereference actual resource once */ \
_resptr->resource_name = (rname); \
_resptr->resource_class = (rclass); \
_resptr->resource_type = (rtype); \
_resptr->resource_size = (rsize); \
_resptr->resource_offset = (roffset); \
_resptr->default_type = (defrtype); \
_resptr->default_addr = (defraddr); \
}

/* Global variables */

typedef struct _vms_item_list {		/* Useful for creating item lists*/
    short	 buffer_length;		/* for system service calls */
    short	 item_code;
    void	*buffer;
    int	*returned_buffer_length;
    } VMS_ItemList;

static XtAppContext app_context;

static Screen  *screen;
static Display *dpy;
static Window root_win;

static Bool     IsColour;

static Pixel state_colour_pxl[MX_STATES];
static char *state_colour_suggestion[MX_STATES] = {
  "YellowGreen",/* Null */
  "Lightgrey",  /* Compatibility - never seen */
  "Black",      /* User mode */
  "Blue",       /* Supervisor mode */
  "Yellow",     /* Executive mode */
  "Orange",     /* Kernel mode */
  "Red"         /* Interrupt mode */
};
static char *state_colour_resource_name[MX_STATES] = {
  "nullColour",     /* Null */
  "compColour",     /* Compatibility - never seen */
  "userColour",     /* User mode */
  "superColour",    /* Supervisor mode */
  "execColour",     /* Executive mode */
  "kernelColour",   /* Kernel mode */
  "intColour"       /* Interrupt mode */
};
static char *state_colour_resource_class[MX_STATES] = {
  "NullColour",     /* Null */
  "CompColour",     /* Compatibility - never seen */
  "UserColour",     /* User mode */
  "SuperColour",    /* Supervisor mode */
  "ExecColour",     /* Executive mode */
  "KernelColour",   /* Kernel mode */
  "IntColour"       /* Interrupt mode */
};

static char  *good_colour_suggestion = "ForestGreen";
static char  *bad_colour_suggestion = "Red";
static Pixel good_colour_pxl;
static Pixel bad_colour_pxl;

static Widget toplevel_widget;         /* Root widget ID of our   */
                                       /* application.            */
static Widget main_window_widget;      /* Root widget ID of main  */
                                       /* MrM fetch               */
static Widget main_popup_widget;       /* The popup exit box */
static Widget contproc_widget;         /* The show proc/cont display */
static Widget generic_message_widget;  /* generic message box     */
static Widget generic_caution_widget;  /* generic caution box     */
static Widget generic_wip_widget;      /* generic WIP box         */
static Widget widget_array[MAX_WIDGETS]; /* Place to keep all other widget IDs*/

static Display*  cpu_graph_disp;
static Display*  state_graph_disp;
static Window    cpu_graph_wind;
static Window    state_graph_wind;
static GC        cpu_graph_gc[MX_STATES];
static GC        state_graph_gc[MX_STATES];

static XmString latin_create;  /* Variables for */
static XmString latin_dismiss; /* compound strings. */
static XmString latin_space;
static XmString latin_zero;
static MrmHierarchy  s_MrmHierarchy;   /* MrM database hierarchy ID */
static MrmType dummy_class;            /* and class variable. */
static char* db_filename_vec[] =          /* MrM heirachy file list. */
               {"PFM$DATA:perf_meter.uid" /* There is only one UID file for */
               };                         /* this application. */
static int db_filename_num =
                (sizeof db_filename_vec / sizeof db_filename_vec [0]);


static XtIntervalId contproc_IntervalId;  /* timeout interval id for contproc */
static int topproc;                       /* index into top process array */
static int topPID;                        /* PID of the top process */
static unsigned int timelast;             /* sys time for contproc update */

static Position contproc_x =  50;    /* Initial location of.. */
static Position contproc_y =  50;    /*  ..cont proc window */
static int contproc_update_secs = 2; /* Contproc updates every 2 sec */
static int contproc_update_int;      /* Contproc updates every .. millisecs */

static Position main_xpos = 856;     /* Initial location of.. */
static Position main_ypos =  36;     /*  ..the main window */
static int main_update_secs = 5;     /* Update secs from resource file */
static int main_update_int;          /* Main window updates every .. millisecs */

static long  last_user;
static long  last_super;
static long  last_exec;
static long  last_kernel;
static long  last_interrupt;
static long  last_compat;
static long  last_spin;
static long  last_idle;

static long  new_null;
static long  new_user;
static long  new_super;
static long  new_exec;
static long  new_kernel;
static long  new_interrupt;
static long  new_compat;
static long  new_idle;
static long  new_spin;

static long total_time,NullTime,UserTime,SupeTime,ExecTime,KernTime,
             InteTime,CompTime,DeltaIO;

/* Used for the contproc routines */
static long NewDirIO, DirIO, DeltaDirIO, LastDeltaDirIO;
static long NewBufIO, BufIO, DeltaBufIO, LastDeltaBufIO;
static long NewPpgcnt, Ppgcnt, DeltaPpgcnt, LastDeltaPpgcnt;
static long NewGpgcnt, Gpgcnt, DeltaGpgcnt, LastDeltaGpgcnt;
static long NewPageflts, Pageflts, DeltaPageflts, LastDeltaPageflts;
static long NewCputim, Cputim, DeltaCputim, LastDeltaCputim;
static long NewWSSize, WSSize, DeltaWSSize, LastDeltaWSSize;
static long NewPgfluse, Pgfluse, DeltaPgfluse, LastDeltaPgfluse;

static int CPUPercent[MX_STATES][101];
static int CPUbase[MX_STATES][101];

static long Device_0_IO;
static long Device_1_IO;
static long Device_2_IO;
static long Device_3_IO;

static long PageFaults;
static long PageIO;
static long FreePages;
static long ModPages;

/* The names and addresses of things that MrM has to bind. The names do */
/* not have to be in alphabetical order */
static MrmRegisterArg reglist[] = {
  {"activate_proc",  (XtPointer) activate_proc},
  {"create_proc",    (XtPointer) create_proc},
  {"cancel_proc",    (XtPointer) cancel_proc},
  {"exposure_proc",  (XtPointer) exposure_proc},
  {"toggle_contproc",(XtPointer) toggle_contproc}
};
static int reglist_num = (sizeof reglist/sizeof reglist[0]);
static char nodename[16];

/*----------------------------------------------------------------------------*/
/* Main Program                                                               */
/*----------------------------------------------------------------------------*/
#ifdef VMS
#  ifdef __DECC
#    pragma message save
#    pragma message disable portable
#  endif

/* The main routine does all the one-time setup and then calls XtAppMainLoop */
PerfMeter(int argc, char *argv[])
MAIN_PROGRAM

#  ifdef __DECC
#    pragma message restore
#  endif
#else

int main(int argc, char *argv[]) /* Standard C version */

#endif
{
    register long i;
    long          timeup_p;
    static char   string[128];
    static unsigned short nlen;
    Pixmap bitmap;
    Pixmap icon_pixmap;
    GC gctemp;
    XGCValues gcvalues;
    XColor exact_colour;
    int x_hot,y_hot;
    unsigned int icon_width,icon_height;
    int istatus;
    struct {
      unsigned short blen, icode;
      char *baddr;
      unsigned short int *rlen;
    } syi_itmlst[] = { {sizeof(nodename)-1,SYI$_SCSNODE,nodename,&nlen},
                       {0, 0, NULL, NULL} };
    union {
      char *string_ptr;
      Bool bool;
    } urs;
    XtResource ursl[MX_STATES+1];
    int arg;
    Arg argl[4];

/*  End of declarations ... */

/*  Get nodename */
    sys$getsyiw(0,0,0,syi_itmlst,0,0,0);
    for (i = nlen; (i >= 0) && (nodename[i] == ' ' || nodename[i] == 0); i--)
    nodename[i] = '\0';

/*  Initialize MrM before initializing the X Toolkit */
    MrmInitialize();

/*  If we had user-defined widgets, we would register them with MrM here */

/*  Initialize the X Toolkit. We get back a top level shell widget */
    toplevel_widget = XtAppInitialize(&app_context,PROGRAM_NAME,
                                      (XrmOptionDescList)NULL,0,
                                      &argc,(String*)argv,
                                      (String*)NULL,
                                      NULL,0);

/*  Get the X server display */
    dpy = XtDisplay(toplevel_widget);

/*  Force synchronization of ALL X calls (use ONLY for DEBUG - very slow) */
/*    XSynchronize(dpy,TRUE);*/

/*  Find screen for this display */
    screen = XtScreen(toplevel_widget);

/*  Find the root window for this screen */
    root_win = XRootWindowOfScreen(screen);

/*  Check if we are on a colour or grey-scale system, or B+w */
    if ( ( (XDefaultVisualOfScreen(screen)->class != PseudoColor)
        && (XDefaultVisualOfScreen(screen)->class != DirectColor)
        && (XDefaultVisualOfScreen(screen)->class != TrueColor)
        && (XDefaultVisualOfScreen(screen)->class != GrayScale) ) ||
        (argc > 1) ) {
      IsColour = FALSE;
    } else {
      IsColour = TRUE;
    }

    if (IsColour) {
/*    Create the colours for positive and negative increments */
      PFM_LOAD_RESOURCE(ursl[0],"goodColour","GoodColour",
                          XtRPixel,sizeof(Pixel),0,XtRString,
                          good_colour_suggestion);
      XtGetApplicationResources(toplevel_widget,&good_colour_pxl,ursl,1,NULL,0);

      PFM_LOAD_RESOURCE(ursl[0],"badColour","badColour",
                          XtRPixel,sizeof(Pixel),0,XtRString,
                          bad_colour_suggestion);
      XtGetApplicationResources(toplevel_widget,&bad_colour_pxl,ursl,1,NULL,0);
    } else {
      bad_colour_pxl = XBlackPixelOfScreen(screen);
      good_colour_pxl = XBlackPixelOfScreen(screen);
    }

/*  Find the colours for different CPU modes */
    for (i = 0; i < MX_STATES; i++) {
      if (IsColour) {
        PFM_LOAD_RESOURCE(ursl[i],state_colour_resource_name[i],
                            state_colour_resource_class[i],
                            XtRPixel,sizeof(Pixel),(i*sizeof(Pixel)),XtRString,
                            state_colour_suggestion[i]);
      } else {
        state_colour_pxl[i] = XBlackPixelOfScreen(screen);
      }
    }
    if (IsColour) XtGetApplicationResources(toplevel_widget,state_colour_pxl,
                                            ursl,MX_STATES,NULL,0);

/*  Set title bar */
    sprintf(string,PROGRAM_TITLE_FORMAT,nodename);
    set_something(toplevel_widget,XmNtitle,(XtArgVal)string);
/*  Set icon name */
    sprintf(string,"%s V%s %s",PROGRAM_NAME,PROGRAM_VERSION,nodename);
    set_something(toplevel_widget,XmNiconName,(XtArgVal)string);

/*  Set icon pixmap */
/*  Create a pixmap from the bitmap format data */
    sprintf(string,"PFM$DATA:PERF_METER_icon.xbm");
    istatus = XReadBitmapFile(dpy,root_win,string,
              &icon_width,&icon_height,&bitmap,&x_hot,&y_hot);
    if (istatus == BitmapSuccess) {
      gcvalues.foreground = XWhitePixelOfScreen(screen);
      gcvalues.background = XBlackPixelOfScreen(screen);
      gctemp = XCreateGC(dpy,root_win,GCForeground|GCBackground,&gcvalues);
      icon_pixmap = XCreatePixmap(dpy,root_win,
        icon_width,icon_height,XDefaultDepthOfScreen(screen));
/*    Now copy 1-plane pixmap (bitmap) to normal pixmap */
      XCopyPlane(dpy,bitmap,icon_pixmap,gctemp,0,0,
                 icon_width,icon_height,0,0,1);
      XFreePixmap(dpy,bitmap);
      XFreeGC(dpy,gctemp);
      set_something(toplevel_widget,XmNiconPixmap,icon_pixmap);
    } else {
      icon_pixmap = 0;
    }

/*  Open the UID files (the output of the UIL compiler) in the hierarchy */
    if (MrmOpenHierarchy(db_filename_num, /* Number of files.            */
                         db_filename_vec, /* Array of file names.        */
                         NULL,            /* Default OS extenstion.      */
                         &s_MrmHierarchy) /* Pointer to returned MrM ID  */
        != MrmSUCCESS) {
      s_error("can't open hierarchy");
      exit(EXIT_FAILURE);
    }

/*  Initialize various data structures */
    init_application();

    initialize_topproc();

/*  Register the items MrM needs to bind for us */
    MrmRegisterNames(reglist, reglist_num);

/*  Go get the main part of the application */
    if (MrmFetchWidget(s_MrmHierarchy, "PERF_METER_MAIN", toplevel_widget,
                       &main_window_widget, &dummy_class) != MrmSUCCESS)
        s_error("can't fetch main window");

/*  Get the sho proc/cont widget */
    if (MrmFetchWidget(s_MrmHierarchy, "CONTPROC_MAIN", main_window_widget,
                       &contproc_widget, &dummy_class) != MrmSUCCESS)
        s_error("can't fetch contproc_widget");

/*  Manage main window,realize everything. Interface appears on the display */
    XtManageChild(main_window_widget);

/*  Get the user specified resources from the default database */

/*  Main window */
    sprintf(string,"%d",main_xpos);
    PFM_LOAD_RESOURCE(ursl[0],"xpos","Xpos",
                        XtRPosition,sizeof(Position),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&main_xpos,ursl,1,NULL,0);
    sprintf(string,"%d",main_ypos);
    PFM_LOAD_RESOURCE(ursl[0],"ypos","Ypos",
                        XtRPosition,sizeof(Position),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&main_ypos,ursl,1,NULL,0);
    set_something(toplevel_widget,XmNx,main_xpos);
    set_something(toplevel_widget,XmNy,main_ypos);

    sprintf(string,"%d",main_update_secs);
    PFM_LOAD_RESOURCE(ursl[0],"UpdatePeriod","UpdatePeriod",
                        XtRInt,sizeof(int),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&main_update_secs,ursl,1,NULL,0);
    main_update_secs = (main_update_secs > 1) ? main_update_secs : 1;
    main_update_int = main_update_secs*1000;

/*  ContProc window */
    sprintf(string,"%d",contproc_x);
    PFM_LOAD_RESOURCE(ursl[0],"contproc_x","ContProc_x",
                        XtRPosition,sizeof(Position),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&contproc_x,ursl,1,NULL,0);
    sprintf(string,"%d",contproc_y);
    PFM_LOAD_RESOURCE(ursl[0],"contproc_y","ContProc_y",
                        XtRPosition,sizeof(Position),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&contproc_y,ursl,1,NULL,0);
    set_something(contproc_widget,XmNx,contproc_x);
    set_something(contproc_widget,XmNy,contproc_y);

    sprintf(string,"%d",contproc_update_secs);
    PFM_LOAD_RESOURCE(ursl[0],"ContProcUpdatePeriod","ContProcUpdatePeriod",
                        XtRInt,sizeof(int),0,XtRString,string);
    XtGetApplicationResources(toplevel_widget,&contproc_update_secs,ursl,1,NULL,0);
    contproc_update_secs = 
           (contproc_update_secs > 1) ? contproc_update_secs : 1;
    contproc_update_int = contproc_update_secs*1000;

    XtRealizeWidget(toplevel_widget);

/*  Get windows and set up GCs for the graphs */
    cpu_graph_wind = XtWindow(widget_array[PFM_CPU_GRAPH]);
    for (i = 0; i < MX_STATES; i++) {
      cpu_graph_gc[i] = MakeGC(cpu_graph_disp,cpu_graph_wind,
                               GXcopy,state_colour_pxl[i]);
    }
    state_graph_wind = XtWindow(widget_array[PFM_STATE_GRAPH]);
    for (i = 0; i < MX_STATES; i++) {
      state_graph_gc[i] = MakeGC(state_graph_disp,state_graph_wind,
                                 GXcopy,state_colour_pxl[i]);
    }

/*  Get the sho proc/cont widget */
    if (MrmFetchWidget(s_MrmHierarchy, "MAIN_POPUP", main_window_widget,
                       &main_popup_widget, &dummy_class) != MrmSUCCESS)
        s_error("can't fetch main_popup_widget");

    /* Hook up the main popup menu */
    XtAddEventHandler (main_window_widget, ButtonPressMask, 0,
			doMainPopupMenu, main_popup_widget);

/*  Add timeout callback that will be called every main_update_secs seconds */
    XtAppAddTimeOut( app_context, main_update_int, update, 0);

/*  Enter the continous event loop ... */
    XtAppMainLoop(app_context);
}

/*----------------------------------------------------------------------------*/
/* One-time initialization of application data structures                     */
/*----------------------------------------------------------------------------*/
void init_application(void)
{
    int k;
/*  End of declarations ... */

    for (k = 0; k < MAX_WIDGETS; k++) widget_array[k] = NULL;
    init_cpu_times();
}

/*----------------------------------------------------------------------------*/
/* Error reporter - All errors are fatal.                                     */
/*----------------------------------------------------------------------------*/
void s_error(char* problem_string)
{
/*  End of declarations ... */
    printf("%s\n", problem_string);
    exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/* Simplified SET VALUE routine to use only when changing a single attribute. */
/* If we need to change more than one, all new values should be put into one  */
/* arglist and we should make one XtSetValues call (MUCH more efficient).     */
/*----------------------------------------------------------------------------*/
void set_something(Widget w, char* resource, XtArgVal value)
{
    Arg al[1];
/*  End of declarations ... */

    XtSetArg(al[0], resource, value);
    XtSetValues(w, al, 1);
}

/*----------------------------------------------------------------------------*/
/* Simplified SET VALUE routine to use only when changing a label widget's    */
/* label string.  Does a Latin1 String and frees the compound string after    */
/* setting label value                                                        */
/*----------------------------------------------------------------------------*/
void set_label(Widget w, char* value)
{
    Arg al[1];
    XmString latin_label;
/*  End of declarations ... */

    latin_label = XmStringCreateSimple(value);
    XtSetArg(al[0], XmNlabelString, latin_label);
    XtSetValues(w, al, 1);
    XmStringFree(latin_label);
}


/*  Set label widget using index in widget array */
void set_label_str (char *value, int w_idx)
{
    Arg al[1];
    XmString CStr;
/*  End of declarations ... */
    CStr = XmStringCreateSimple(value);
    XtSetArg(al[0], XmNlabelString, CStr);
    XtSetValues(widget_array[w_idx], al, 1);
    XmStringFree(CStr);
}

/*----------------------------------------------------------------------------*/
/* process_num() handles all of the repetitive stuff needed to track          */
/* what's changed from update to update.  We also update the widget           */
/* label if something has changed since the last time.                        */
/*----------------------------------------------------------------------------*/
void process_num(long *NewValue, long *Value, long *DeltaValue,
		 long *LastDeltaValue, int ValueIdx, int DeltaIdx,
		 char *ValueFmt, char *DeltaFmt)
{
    char label_buffer[15];
    int first_time;
/*  End of declarations ... */

    *DeltaValue = *NewValue - *Value;
    first_time = !(*Value);
    if (*DeltaValue) {
      *Value = *NewValue;
      sprintf (label_buffer, ValueFmt, *Value);
      set_label_str( label_buffer, ValueIdx );
    }
    if ( *DeltaValue != *LastDeltaValue ) {
      if (*DeltaValue == 0 || first_time ) {
        label_buffer[0] = ' ';
        label_buffer[1] = '\0';
      } else {
        sprintf (label_buffer, DeltaFmt, *DeltaValue);
      }
      set_label_str( label_buffer, DeltaIdx );

      *LastDeltaValue = *DeltaValue;
    }
}

/*----------------------------------------------------------------------------*/
/* Simplified GET VALUE routine to use only when retrieving a SINGLE attribute*/
/*----------------------------------------------------------------------------*/
void get_something(Widget w, char* resource, XtArgVal value)
{
    Arg al[1];
/*  End of declarations ... */

    XtSetArg(al[0], resource, value);
    XtGetValues(w, al, 1);
}

/*----------------------------------------------------------------------------*/
/* All widgets that are created call back to this procedure. We just log the  */
/* ID in the global array.                                                    */
/*----------------------------------------------------------------------------*/
void create_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s)
{
    int widget_num = *(int *)client_data;
/*  End of declarations ... */

    widget_array[widget_num] = w;

    switch ( widget_num ) {
      case PFM_CPU_GRAPH:
        cpu_graph_disp = XtDisplay(w);
        break;
      case PFM_STATE_GRAPH:
        state_graph_disp = XtDisplay(w);
        break;
      case PFM_DEVICE_0_IOCOUNT_NAME:
        update_disk_name_label(w, 0);
        break;
      case PFM_DEVICE_1_IOCOUNT_NAME:
        update_disk_name_label(w, 1);
        break;
      case PFM_DEVICE_2_IOCOUNT_NAME:
        update_disk_name_label(w, 2);
        break;
      case PFM_DEVICE_3_IOCOUNT_NAME:
        update_disk_name_label(w, 3);
        break;
        default:
      break;
    }

}

/*----------------------------------------------------------------------------*/
/* exposure_proc() handles exposure events for the cpu histogram and          */
/* state graph.  This is called whenever the window receives an               */
/* exposure event.   We need to redraw the graphics  ourselves.               */
/*----------------------------------------------------------------------------*/
void exposure_proc (Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s)
{
    int i;
    int ix;
    int index;
/*  End of declarations ... */

    index = *(int *)client_data;
    switch ( index ) {
      case PFM_CPU_GRAPH:
        for (ix = 0; ix < 100; ix++) {
          for (i = 0; i < MX_STATES; i++) {
            if (i == 0) {
              XClearArea(cpu_graph_disp,cpu_graph_wind,ix,0,1,100,0);
            } else {
              if (CPUPercent[i][ix]) {
                XFillRectangle(cpu_graph_disp,cpu_graph_wind,cpu_graph_gc[i],
                ix,100-CPUPercent[i][ix]-CPUbase[i][ix],1,CPUPercent[i][ix]);
              }
            }
          }
        }
        break;
      case PFM_STATE_GRAPH:
        DrawStateGraph();
        break;
    }
}

/*----------------------------------------------------------------------------*/
/* All push buttons in this application call back to this routine.  We        */
/* use the client_data to tell us what widget it is, then react accordingly.  */
/*----------------------------------------------------------------------------*/
void activate_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s)
{
    int widget_num = *(int *)client_data; /* Convert to widget number. */
/*  End of declarations ... */

    switch ( widget_num ) {
      case PFM_EXIT_BUTTON: exit(EXIT_SUCCESS);
      default: break;
    }
}

/*----------------------------------------------------------------------------*/
/* Not used, apparently                                                       */
/*----------------------------------------------------------------------------*/
void cancel_proc(Widget w, XtPointer client_data, XmAnyCallbackStruct *cb_s)
{
    int widget_num = *(int *)client_data; /* Convert to widget number. */
/*  End of declarations ... */

    switch ( widget_num ) {
      default: break;
    }
}

/*----------------------------------------------------------------------------*/
/* Update the performance meter main statistics - called as a timer callback  */
/*----------------------------------------------------------------------------*/
void update( XtPointer data, XtIntervalId *id )
{
    register int i,j;
    static long NewNullTime,NewUserTime,NewSupeTime,NewExecTime,
                NewKernTime,NewInteTime,NewCompTime,
                NewDevice_0_IO, DeltaDevice_0_IO, LastDeltaDevice_0_IO,
                NewDevice_1_IO, DeltaDevice_1_IO, LastDeltaDevice_1_IO,
                NewDevice_2_IO, DeltaDevice_2_IO, LastDeltaDevice_2_IO,
                NewDevice_3_IO, DeltaDevice_3_IO, LastDeltaDevice_3_IO,
                LastDeltaPageFaults, LastDeltaPageIO,
                LastDeltaFreePages, LastDeltaModPages;

    long NewPageFaults,DeltaPageFaults,NewPageIO,DeltaPageIO,
         NewFreePages, DeltaFreePages,
         NewModPages, DeltaModPages, NewDispFlag;
    int ix;
    long CPUtime[MX_STATES];
    char *label_topproc;
    static char label_buffer[32];
    XmString cs;
/*  End of declarations ... */

    get_cpu_times(&NewNullTime,&NewUserTime,&NewSupeTime,&NewExecTime,
                  &NewKernTime,&NewInteTime,&NewCompTime);
    NewPageFaults = CURRENT_FAULTS;
    NewPageIO     = CURRENT_PAGEIO;
    NewFreePages  = CURRENT_FREEPAGES;
    NewModPages   = CURRENT_MODPAGES;

    total_time = NewNullTime + NewUserTime + NewSupeTime + NewExecTime +
         NewKernTime + NewInteTime + NewCompTime;
    if (total_time > 0) {
      NullTime = (NewNullTime * 100) / total_time;
      UserTime = (NewUserTime * 100) / total_time;
      SupeTime = (NewSupeTime * 100) / total_time;
      ExecTime = (NewExecTime * 100) / total_time;
      KernTime = (NewKernTime * 100) / total_time;
      InteTime = (NewInteTime * 100) / total_time;
      CompTime = (NewCompTime * 100) / total_time;
    }
    else {
      NullTime = NewNullTime;
      UserTime = NewUserTime;
      SupeTime = NewSupeTime;
      ExecTime = NewExecTime;
      KernTime = NewKernTime;
      InteTime = NewInteTime;
      CompTime = NewCompTime;
    }

    CPUtime[0] = NullTime;
    CPUtime[1] = CompTime;
    CPUtime[2] = UserTime;
    CPUtime[3] = SupeTime;
    CPUtime[4] = ExecTime;
    CPUtime[5] = KernTime;
    CPUtime[6] = InteTime;

    for (ix = 0; ix < 100; ix++) {
      for (i = 0; i < MX_STATES; i++) {
        CPUbase[i][100] = 0;
        CPUPercent[i][100] = 100;
        for (j = 0; j < MX_STATES; j++) {
          if (j != i) CPUPercent[i][100] -= CPUtime[j];
        }
        if (i == 0) {
          XClearArea(cpu_graph_disp,cpu_graph_wind,ix,0,1,100,0);
        } else {
          if (i > 1) CPUbase[i][100] = CPUtime[i-1]+CPUbase[i-1][100];
          CPUPercent[i][ix] = CPUPercent[i][ix+1];
          CPUbase[i][ix] = CPUbase[i][ix+1];
          if (CPUPercent[i][ix]) {
            XFillRectangle(cpu_graph_disp,cpu_graph_wind,cpu_graph_gc[i],
            ix,100-CPUPercent[i][ix]-CPUbase[i][ix],1,CPUPercent[i][ix]);
          }
        }
      }
    }

/*  Get latest top process */
    topproc = get_topproc();
    label_topproc = format_topproc(topproc);
    set_label_str (label_topproc, PFM_CONTPROC_TOGGLE);

    DrawStateGraph();

    if ((widget_array[PFM_DEVICE_0_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_0_IOCOUNT_DELTA] != NULL)) {
      get_disk_opcnt(0, &NewDevice_0_IO);
      process_num (&NewDevice_0_IO, &Device_0_IO, &DeltaDevice_0_IO, &LastDeltaDevice_0_IO,
                   PFM_DEVICE_0_IOCOUNT_TOTAL, PFM_DEVICE_0_IOCOUNT_DELTA, " %10d", "  %4d");
    }


    if ((widget_array[PFM_DEVICE_1_IOCOUNT_TOTAL] != NULL) &&
        (widget_array[PFM_DEVICE_1_IOCOUNT_DELTA] != NULL)) {
      get_disk_opcnt(1, &NewDevice_1_IO);
      process_num (&NewDevice_1_IO, &Device_1_IO, &DeltaDevice_1_IO, &LastDeltaDevice_1_IO,
                   PFM_DEVICE_1_IOCOUNT_TOTAL, PFM_DEVICE_1_IOCOUNT_DELTA, " %10d", "  %4d");
    }

    if ((widget_array[PFM_DEVICE_2_IOCOUNT_TOTAL] != NULL) &&
        (widget_array[PFM_DEVICE_2_IOCOUNT_DELTA] != NULL)) {
      get_disk_opcnt(2, &NewDevice_2_IO);
      process_num (&NewDevice_2_IO, &Device_2_IO, &DeltaDevice_2_IO, &LastDeltaDevice_2_IO,
                   PFM_DEVICE_2_IOCOUNT_TOTAL, PFM_DEVICE_2_IOCOUNT_DELTA, " %10d", "  %4d");
    }

    if ((widget_array[PFM_DEVICE_3_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_3_IOCOUNT_DELTA] != NULL)) {
      get_disk_opcnt(3, &NewDevice_3_IO);
      process_num (&NewDevice_3_IO, &Device_3_IO, &DeltaDevice_3_IO, &LastDeltaDevice_3_IO,
                   PFM_DEVICE_3_IOCOUNT_TOTAL, PFM_DEVICE_3_IOCOUNT_DELTA, " %10d", "  %4d");
    }

    process_num (&NewPageFaults, &PageFaults, &DeltaPageFaults, &LastDeltaPageFaults,
                 PFM_PAGEFAULTS_TOTAL, PFM_PAGEFAULTS_DELTA, " %10d", "  %4d");
    set_something(widget_array[PFM_PAGEFAULTS_DELTA],XmNforeground,
                  (DeltaPageFaults > 0) ? bad_colour_pxl : good_colour_pxl );

    process_num (&NewPageIO, &PageIO, &DeltaPageIO, &LastDeltaPageIO,
                 PFM_PAGEIO_TOTAL, PFM_PAGEIO_DELTA, " %10d", "  %4d");
    set_something(widget_array[PFM_PAGEIO_DELTA],XmNforeground,
                  (DeltaPageIO > 0) ? bad_colour_pxl : good_colour_pxl );

    process_num (&NewFreePages, &FreePages, &DeltaFreePages, &LastDeltaFreePages,
                 PFM_FREEPAGES_TOTAL, PFM_FREEPAGES_DELTA, " %10d", "  %4d");
    set_something(widget_array[PFM_FREEPAGES_DELTA],XmNforeground,
                  (DeltaFreePages < 0) ? bad_colour_pxl : good_colour_pxl );

    process_num (&NewModPages, &ModPages, &DeltaModPages, &LastDeltaModPages,
                 PFM_MODPAGES_TOTAL, PFM_MODPAGES_DELTA, " %10d", "  %4d");
    set_something(widget_array[PFM_MODPAGES_DELTA],XmNforeground,
                    (DeltaModPages > 0) ? bad_colour_pxl : good_colour_pxl );

    XtAppAddTimeOut( app_context, main_update_int, update, 0);
}

/*----------------------------------------------------------------------------*/
/* Draw the cpu state graph                                                   */
/*----------------------------------------------------------------------------*/
void DrawStateGraph(void)
{
    int i;
/*  End of declarations ... */

    XClearArea(state_graph_disp,state_graph_wind,NullTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[0],
                   NullTimeX,100-NullTime,7,NullTime);
    XClearArea(state_graph_disp,state_graph_wind,CompTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[1],
                   CompTimeX,100-CompTime,7,CompTime);
    XClearArea(state_graph_disp,state_graph_wind,UserTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[2],
                   UserTimeX,100-UserTime,7,UserTime);
    XClearArea(state_graph_disp,state_graph_wind,SupeTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[3],
                   SupeTimeX,100-SupeTime,7,SupeTime);
    XClearArea(state_graph_disp,state_graph_wind,ExecTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[4],
                   ExecTimeX,100-ExecTime,7,ExecTime);
    XClearArea(state_graph_disp,state_graph_wind,KernTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[5],
                   KernTimeX,100-KernTime,7,KernTime);
    XClearArea(state_graph_disp,state_graph_wind,InteTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc[6],
                   InteTimeX,100-InteTime,7,InteTime);
}

/*----------------------------------------------------------------------------*/
/* Create a Graphics Context with appropriate function and foreground         */
/*----------------------------------------------------------------------------*/
GC MakeGC(Display* disp, Window wind, long fun, Pixel foreground)
{
    XColor colour,exact_colour;
    XGCValues gcvalues;
/*  End of declarations ... */

    gcvalues.foreground = foreground;;
    gcvalues.function = fun;
    return(XCreateGC(disp,wind,GCFunction|GCForeground,&gcvalues));
}

/*----------------------------------------------------------------------------*/
/* Get first set of CPU stats by setting up a kernel mode routine             */
/*----------------------------------------------------------------------------*/
void init_cpu_times()
{
/*  End of declarations ... */

    last_user = last_super = last_exec = last_kernel = last_interrupt =
    last_compat = last_spin = last_idle = 0;

#ifdef __ALPHA
    sys$cmkrnl(&get_initial_kernelly_stuff, NULL);
#else
    get_initial_kernelly_stuff();
#endif

    last_interrupt -= last_idle;
}

/*----------------------------------------------------------------------------*/
/* This is called at Kernel mode to actually look up the information          */
/*----------------------------------------------------------------------------*/
void get_initial_kernelly_stuff()
{
    register   i, j, k;
    long      *PMS$GL_KERNEL;
/*  End of declarations ... */

    for (k = SMP$GL_ACTIVE_CPUS, j = 0; k != 0; j++, k = k >> 1) {
      if ((k & 01) == 0) continue;
#ifdef __ALPHA
      PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$q_kernel;
      last_compat = 0;
#else
      PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$l_kernel;
      last_compat += PMS$GL_KERNEL[C_TIME];
#endif
      last_kernel += PMS$GL_KERNEL[K_TIME];
      last_exec += PMS$GL_KERNEL[E_TIME];
      last_super += PMS$GL_KERNEL[S_TIME];
      last_user += PMS$GL_KERNEL[U_TIME];
      last_interrupt += PMS$GL_KERNEL[I_TIME];
      last_spin += PMS$GL_KERNEL[SPIN_TIME];
/*    here is where the null CPU time is collected.  */
      last_idle += PMS$GL_KERNEL[IDLE_TIME];
    }
}

/*----------------------------------------------------------------------------*/
/* Get next set of CPU stats by setting up a kernel mode routine              */
/*----------------------------------------------------------------------------*/
void get_cpu_times(long* null, long* user, long* super, long* exec,
                          long* kernel, long* interrupt, long* compat)
{
    register long  new_abstime;
    register long  totaltime;
/*  End of declarations ... */

    new_user = new_super = new_exec = new_kernel = new_interrupt =
    new_compat = new_spin = new_idle = 0;

#ifdef __ALPHA
    sys$cmkrnl(&get_new_kernelly_stuff, NULL);
#else
    get_new_kernelly_stuff();
#endif

    new_interrupt -= new_idle;

    *user = new_user - last_user; last_user = new_user;
    *super = new_super - last_super; last_super = new_super;
    *exec = new_exec - last_exec; last_exec = new_exec;
    *kernel = new_kernel - last_kernel; last_kernel = new_kernel;
    *interrupt = new_interrupt - last_interrupt; last_interrupt = new_interrupt;
    *compat = new_compat - last_compat; last_compat = new_compat;

    *null = new_idle - last_idle; last_idle = new_idle;
}

/*----------------------------------------------------------------------------*/
/* This is called at Kernel mode to actually look up the information          */
/*----------------------------------------------------------------------------*/
void get_new_kernelly_stuff()
{
    register   i, j, k;
    long      *PMS$GL_KERNEL;

/*  End of declarations ... */

    for (k = SMP$GL_ACTIVE_CPUS, j = 0; k != 0; j++, k = k >> 1) {
      if ((k & 01) == 0) continue;
#ifdef __ALPHA
        PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$q_kernel;
        new_compat = 0;
#else
        PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$l_kernel;
        new_compat += PMS$GL_KERNEL[C_TIME];
#endif
        new_kernel += PMS$GL_KERNEL[K_TIME];
        new_exec += PMS$GL_KERNEL[E_TIME];
        new_super += PMS$GL_KERNEL[S_TIME];
        new_user += PMS$GL_KERNEL[U_TIME];
        new_interrupt += PMS$GL_KERNEL[I_TIME];
        new_spin += PMS$GL_KERNEL[SPIN_TIME];
        new_idle += PMS$GL_KERNEL[IDLE_TIME];
    }
}

/*----------------------------------------------------------------------------*/
/* Get disk IO operations                                                     */
/*----------------------------------------------------------------------------*/
void get_disk_opcnt(register long devnum, register long* count)
{
    static struct {
      unsigned short  blen;
      unsigned short  icode;
      long*           buff;
      unsigned short* rlen;
    } dvi_itmlst[] = { {sizeof(int), DVI$_OPCNT, NULL, NULL},
                       {0, 0, NULL, NULL} };
    static struct {
      unsigned short  istat;
      unsigned short  bytes;
      long            fill;
    } iosb;
    static char devname[64];
    $DESCRIPTOR(dev,devname);
    long   status;
/*  End of declarations ... */

    sprintf(devname, "PFM$DEVICE_%d_NAME", devnum);
    dvi_itmlst[0].buff = count;
    *count = -1;
    dev.dsc$w_length = strlen(devname);
    status = sys$getdviw(0,0,&dev,dvi_itmlst,&iosb,0,0,0);
    if (status == SS$_IVDEVNAM) {
      *count = 0;
      if (devnum == 0) {
        $DESCRIPTOR(sysdev, "SYS$SYSDEVICE:");
        sys$getdviw(0,0,&sysdev,dvi_itmlst,&iosb,0,0,0);
      }
    }
}

/*----------------------------------------------------------------------------*/
/* Set title for monitored disk IO                                            */
/*----------------------------------------------------------------------------*/
void update_disk_name_label(Widget w, long devnum)
{
    char   devname[64];
    char*  title;
/*  End of declarations ... */

    sprintf(devname, "PFM$DEVICE_%d_TITLE", devnum);
    title = getenv(devname);
    if (NULL == title) {
      sprintf(devname, "PFM$DEVICE_%d_NAME", devnum);
      title = getenv(devname);
      if ((NULL == title) && (0 == devnum)) title = "SYS$SYSDEVICE:";
    }
    if (title != NULL) set_label(w,title);
}

/*----------------------------------------------------------------------------*/
/* This callback is invoked when user clicks on the `contproc' toggle         */
/*----------------------------------------------------------------------------*/
void toggle_contproc (Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *toggle)
{
/*  End of declarations ... */
    contproc_off();                  /* Turn off any existing display */
    if (toggle->set) contproc_on (); /* Turn display on if needed */
}

/*----------------------------------------------------------------------------*/
/* Procecures to turn 'show proc/cont' display off, unmanaging the dialog box */
/*----------------------------------------------------------------------------*/
void contproc_off(void)
{
/*  End of declarations ... */

    if ( contproc_IntervalId != 0) {
      XtRemoveTimeOut(contproc_IntervalId);
      contproc_IntervalId = 0;
    }
    if (contproc_widget!=0) {
      /* Save the window position so we can put it back exactly */
      /* where the use last moved it to, otherwise, it pops up  */
      /* all over the place. */
      if (XtIsManaged (contproc_widget)) {
        get_something (XtParent(contproc_widget), XmNx, (XtArgVal)&contproc_x);
        get_something (XtParent(contproc_widget), XmNy, (XtArgVal)&contproc_y);
        XtUnmanageChild (contproc_widget);
      }
    }
}

/*----------------------------------------------------------------------------*/
/* contrproc_on()  initializes the 'show proc/cont' display.  That is,        */
/* it looks up the static information for this process, zeros our counters,   */
/* realizes the window, and makes the initial call to the contproc_update     */
/* procedure.                                                                 */
/*----------------------------------------------------------------------------*/
void contproc_on(void)
{
    char text[100];
    int userlen, prcstrt, status;
    unsigned int timequad[2];

    static int pid;
    static char username[12], prcnam[15];
    static VMS_ItemList JPI_static_itemlist[4] = {
        sizeof(pid),      JPI$_PID,      &pid, 0,
        sizeof(username), JPI$_USERNAME, &username, 0,
        sizeof(prcnam),   JPI$_PRCNAM,   &prcnam, 0,
        0, 0, 0, 0};
/*  End of declarations ... */

/*  Zero out `previous' values so process_num doesn't give silly deltas */
    DirIO = 0;
    BufIO = 0;
    Ppgcnt = 0;
    Gpgcnt = 0;
    Pageflts = 0;
    Cputim = 0;
    WSSize = 0;
    Pgfluse = 0;

    sys$gettim(&timequad[0]);		/* To calc % cpu used */
    timelast = timequad[0];

    /* Remember the PID of the top user - we monitor this same process */
    /* as long the this window is active */
    topPID = topproc_getPID (topproc);

    status = sys$getjpiw (0, &topPID, 0, &JPI_static_itemlist, 0,0,0);
    if (status != SS$_NORMAL)
        {
        contproc_jpi_status (status);
        return;
        }

    /* Trim trailing spaces from username */
    for (userlen=0; userlen<=sizeof(username) && username[userlen]!=' '; userlen++);

    /* Trim leading spaces from process name */
    for (prcstrt=0; prcstrt<=sizeof(prcnam) && prcnam[prcstrt]==' '; prcstrt++);

    sprintf (text, "PID: %08X  User: %.*s  Process: %.*s\0",
	pid, userlen, username,  (sizeof(prcnam)-prcstrt), (prcnam+prcstrt));
    set_label_str (text, PFM_HEADER1);

    /* Put the window back where it was last time */
    XtManageChild (contproc_widget);
    set_something (XtParent(contproc_widget), XmNx, contproc_x);
    set_something (XtParent(contproc_widget), XmNy, contproc_y);
    XtRealizeWidget (contproc_widget);

    DirIO = LastDeltaDirIO = BufIO = LastDeltaBufIO = Ppgcnt =
        LastDeltaPpgcnt = Gpgcnt = LastDeltaGpgcnt = 
        Pageflts = LastDeltaPageflts = Cputim = LastDeltaCputim = 
	WSSize = LastDeltaWSSize = Pgfluse = LastDeltaPgfluse = 0;

    update_contproc(NULL, NULL);		/* get things rolling */
}

/*----------------------------------------------------------------------------*/
/* Timer callback to update stats for cont. monitored `top CPU' process       */
/*----------------------------------------------------------------------------*/
void update_contproc( XtPointer data, XtIntervalId *id )
{
    int status, tmp;
    char CputimStr[12], CpupctStr[6];
    static unsigned int timequad[2], timediff;

    static int imaglen, pagfilcnt, pgflquota;
    static char imagname[80];
    static VMS_ItemList JPI_contproc_itemlist[10] = {
        sizeof(NewDirIO),    JPI$_DIRIO,    &NewDirIO, 0,
        sizeof(NewBufIO),    JPI$_BUFIO,    &NewBufIO, 0,
        sizeof(NewCputim),   JPI$_CPUTIM,   &NewCputim, 0,
        sizeof(NewPpgcnt),   JPI$_PPGCNT,   &NewPpgcnt, 0,
        sizeof(NewGpgcnt),   JPI$_GPGCNT,   &NewGpgcnt, 0,
	sizeof(pagfilcnt),   JPI$_PAGFILCNT,&pagfilcnt, 0,
	sizeof(pgflquota),   JPI$_PGFLQUOTA,&pgflquota, 0,
        sizeof(NewPageflts), JPI$_PAGEFLTS, &NewPageflts, 0,
        sizeof(imagname),    JPI$_IMAGNAME, &imagname, &imaglen,
        0, 0, 0, 0};
/*  End of declarations ... */

    /* We need the change in system time to calc the % cpu used */
    sys$gettim (&timequad[0]);
    timediff = timequad[0] - timelast;
    timelast = timequad[0];

    /* Now get the stats for the top process */
    status = sys$getjpiw (0, &topPID, 0, &JPI_contproc_itemlist, 0,0,0);
    if (status != SS$_NORMAL)
        {
        contproc_jpi_status (status);
        return;
        }

    /* It makes more sense to me to see page file use */
    /* instead of page file quota remaining. */
    NewPgfluse = pgflquota - pagfilcnt;

    /* We show process and group page counts, but it's also useful */
    /* to see the totaled working set size */
    NewWSSize = NewPpgcnt + NewGpgcnt;


    process_num (&NewPpgcnt, &Ppgcnt, &DeltaPpgcnt, &LastDeltaPpgcnt,   
		 PFM_PPGCNT_TOTAL, PFM_PPGCNT_DELTA,  "%8d", "%8d");
    process_num (&NewGpgcnt, &Gpgcnt, &DeltaGpgcnt, &LastDeltaGpgcnt,   
		 PFM_GPGCNT_TOTAL, PFM_GPGCNT_DELTA,  "%8d", "%8d");
    process_num (&NewWSSize, &WSSize, &DeltaWSSize, &LastDeltaWSSize,   
		 PFM_WSSIZE_TOTAL, PFM_WSSIZE_DELTA,  "%8d", "%8d");
    process_num (&NewPageflts, &Pageflts, &DeltaPageflts, &LastDeltaPageflts, 
		 PFM_PGFAULTS_TOTAL, PFM_PGFAULTS_DELTA,"%8d", "%8d");
    process_num (&NewPgfluse, &Pgfluse, &DeltaPgfluse, &LastDeltaPgfluse, 
		 PFM_PGFLUSE_TOTAL, PFM_PGFLUSE_DELTA,"%8d", "%8d");
    process_num (&NewDirIO, &DirIO, &DeltaDirIO, &LastDeltaDirIO,    
		 PFM_DIRIO_TOTAL, PFM_DIRIO_DELTA,   "%8d", "%8d");
    process_num (&NewBufIO, &BufIO, &DeltaBufIO, &LastDeltaBufIO,
    		 PFM_BUFIO_TOTAL, PFM_BUFIO_DELTA, "%8d", "%8d");

    format_cputime (NewCputim, CputimStr);
    set_label_str (CputimStr, PFM_CPUTIME_TOTAL);

    if ( Cputim ) {
      DeltaCputim = NewCputim - Cputim;
    } else {
      DeltaCputim = 0;
    }
    Cputim = NewCputim;
    LastDeltaCputim = DeltaCputim;
    tmp = timediff / 100000;
    sprintf (CpupctStr, " %3d%%\0", (DeltaCputim*100+50) / (tmp ? tmp : 1) );
    set_label_str (CpupctStr, PFM_CPUTIME_DELTA);

    imagname[imaglen] = '\0';
    set_label_str (imagname, PFM_HEADER2);
             
    contproc_IntervalId =
      XtAppAddTimeOut(app_context, contproc_update_int, update_contproc, 0);
}

/*----------------------------------------------------------------------------*/
/* Find state of monitored process, adjust monitoring accordingly             */
/*----------------------------------------------------------------------------*/
void contproc_jpi_status (int status)
{
    static char text[35];
/*  End of declarations ... */

    switch (status) {
        case SS$_SUSPENDED:
            set_label_str ("(Suspended)", PFM_HEADER2);
            contproc_IntervalId = XtAppAddTimeOut(app_context,
              2*contproc_update_int, update_contproc, 0);
            break;
        case SS$_NONEXPR:
            set_label_str ("Nonexistant process", PFM_HEADER1);
            set_label_str ("(display update stopped)", PFM_HEADER2);
            contproc_IntervalId = 0;
            break;
        default:
            sprintf (text, "Error in $GETJPI, stat=%d", status);
            set_label_str (text, PFM_HEADER1);
            set_label_str ("(display update stopped)", PFM_HEADER2);
            contproc_IntervalId = 0;
            break;
    }
}

/*----------------------------------------------------------------------------*/
/* doMainPopupMenu() is the callback whenever a button is clicked in          */
/* one of the graph displays.   All we do here is display a popup menu        */
/* which contains the option to exit.                                         */
/*----------------------------------------------------------------------------*/
void doMainPopupMenu(Widget wdgt,XtPointer client_data,
                     XEvent *event, Boolean *propogate)
{
    Widget popup;
/*  End of declarations ... */

/*  This makes our popup menu visible */
    popup = (Widget)client_data;
    XmMenuPosition (popup, (XButtonPressedEvent *)event);
    XtManageChild(popup);
    XtRealizeWidget(popup);
}

/*----------------------------------------------------------------------------*/
/* format_cputime() takes an integer (t) which represents a time value        */
/* in 10 millisecond ticks, and returns a string (*s) which contains          */
/* the time formatted as hh:mm:ss.hh                                          */
/*----------------------------------------------------------------------------*/
void format_cputime (int t, char *s)
{
    int hour, min, sec, hund, tmp;
/*  End of declarations ... */

    hour = t / 360000;
    tmp = t - hour*360000;
    min = tmp / 6000;
    tmp = tmp - min*6000;
    sec = tmp / 100;
    hund = tmp - sec*100;

    sprintf (s, "%02d:%02d:%02d.%02d\0", hour, min, sec, hund);
}
